# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
import re

import requests

from .heuristics import Part, parts_heuristics

LOG = logging.getLogger(__name__)


# S3 does not support multiple ranges
class S3Streamer:
    def __init__(self, url, request, parts, headers, **kwargs):
        self.url = url
        self.parts = parts
        self.request = request
        self.headers = dict(**headers)
        self.kwargs = kwargs

    def __call__(self, chunk_size):
        # See https://docs.aws.amazon.com/AmazonS3/latest/API/API_GetObject.html

        headers = dict(**self.headers)
        # TODO: add assertions

        for i, part in enumerate(self.parts):
            if i == 0:
                request = self.request
            else:
                offset, length = part
                headers["range"] = f"bytes={offset}-{offset+length-1}"
                request = requests.get(
                    self.url,
                    stream=True,
                    headers=headers,
                    **self.kwargs,
                )
                try:
                    request.raise_for_status()
                except Exception:
                    LOG.error("URL %s: %s", self.url, request.text)
                    raise

            header = request.headers
            bytes = header["content-range"]
            LOG.debug("HEADERS %s", header)
            m = re.match(r"^bytes (\d+)d?-(\d+)d?/(\d+)d?$", bytes)
            assert m, header
            start, end, total = int(m.group(1)), int(m.group(2)), int(m.group(3))

            assert end >= start
            assert start < total
            assert end < total

            assert start == part.offset, (bytes, part)
            # (end + 1 == total) means that we overshoot the end of the file,
            # this happens when we round transfer blocks
            assert (end == part.offset + part.length - 1) or (end + 1 == total), (
                bytes,
                part,
            )

            yield from request.iter_content(chunk_size)


class MultiPartStreamer:
    def __init__(self, url, request, parts, boundary, **kwargs):
        self.request = request
        self.size = int(request.headers["content-length"])
        self.encoding = "utf-8"
        self.parts = parts
        self.boundary = boundary

    def __call__(self, chunk_size):
        from email.parser import HeaderParser

        from requests.structures import CaseInsensitiveDict

        header_parser = HeaderParser()
        marker = f"--{self.boundary}\r\n".encode(self.encoding)
        end_header = b"\r\n\r\n"
        end_data = b"\r\n"

        end_of_input = f"--{self.boundary}--\r\n".encode(self.encoding)

        if chunk_size < len(end_data):
            chunk_size = len(end_data)

        iter_content = self.request.iter_content(chunk_size)
        chunk = next(iter_content)

        # Some servers start with \r\n
        if chunk[:2] == end_data:
            chunk = chunk[2:]

        LOG.debug("MARKER %s", marker)
        part = 0
        while True:
            while len(chunk) < max(len(marker), len(end_of_input)):
                more = next(iter_content)
                assert more is not None
                chunk += more

            if chunk.find(end_of_input) == 0:
                assert part == len(self.parts)
                break

            pos = chunk.find(marker)
            assert pos == 0, (pos, chunk)

            chunk = chunk[pos + len(marker) :]
            while True:
                pos = chunk.find(end_header)
                if pos != -1:
                    break
                more = next(iter_content)
                assert more is not None
                chunk += more
                assert len(chunk) < 1024 * 1024

            pos += len(end_header)
            header = chunk[:pos].decode(self.encoding)
            header = CaseInsensitiveDict(header_parser.parsestr(header))
            chunk = chunk[pos:]
            # kind = header["content-type"]
            bytes = header["content-range"]
            LOG.debug("HEADERS %s", header)
            m = re.match(r"^bytes (\d+)d?-(\d+)d?/(\d+)d?$", bytes)
            assert m, header
            start, end, total = int(m.group(1)), int(m.group(2)), int(m.group(3))

            assert end >= start
            assert start < total
            assert end < total

            size = end - start + 1

            assert start == self.parts[part].offset
            # (end + 1 == total) means that we overshoot the end of the file,
            # this happens when we round transfer blocks
            assert (end == self.parts[part].offset + self.parts[part].length - 1) or (
                end + 1 == total
            ), (bytes, self.parts[part])

            while size > 0:
                if len(chunk) >= size:
                    yield chunk[:size]
                    chunk = chunk[size:]
                    size = 0
                else:
                    yield chunk
                    size -= len(chunk)
                    chunk = next(iter_content)

            assert chunk.find(end_data) == 0
            chunk = chunk[len(end_data) :]
            part += 1


class DecodeMultipart:
    def __init__(self, url, request, parts, **kwargs):
        self.request = request
        assert request.status_code == 206, request.status_code

        content_type = request.headers["content-type"]

        if content_type.startswith("multipart/byteranges; boundary="):
            _, boundary = content_type.split("=")
            # print("******  MULTI-PART supported by server", url)
            self.streamer = MultiPartStreamer(url, request, parts, boundary, **kwargs)
        else:
            # print("******  MULTI-PART *NOT* supported by server", url)
            self.streamer = S3Streamer(url, request, parts, **kwargs)

    def __call__(self, chunk_size):
        return self.streamer(chunk_size)


class PartFilter:
    def __init__(self, parts, positions=None):
        self.parts = parts

        if positions is None:
            positions = [x.offset for x in parts]
        self.positions = positions

        assert len(self.parts) == len(self.positions)

    def __call__(self, streamer):
        def execute(chunk_size):
            stream = streamer(chunk_size)
            chunk = next(stream)
            pos = 0
            for (_, length), offset in zip(self.parts, self.positions):
                offset -= pos

                while offset > len(chunk):
                    pos += len(chunk)
                    offset -= len(chunk)
                    chunk = next(stream)
                    assert chunk

                chunk = chunk[offset:]
                pos += offset
                size = length
                while size > 0:
                    if len(chunk) >= size:
                        yield chunk[:size]
                        chunk = chunk[size:]
                        pos += size
                        size = 0
                    else:
                        yield chunk
                        size -= len(chunk)
                        pos += len(chunk)
                        chunk = next(stream)

            # Drain stream, so we don't created error messages in the server's logs
            while True:
                try:
                    next(stream)
                except StopIteration:
                    break

        return execute


def compress_parts(parts):
    last = -1
    result = []
    # Compress and check
    for offset, length in parts:
        assert offset >= 0 and length > 0
        assert offset >= last, (
            f"Offsets and lengths must be in order, and not overlapping:"
            f" offset={offset}, end of previous part={last}"
        )
        if offset == last:
            # Compress
            offset, prev_length = result.pop()
            length += prev_length

        result.append((offset, length))
        last = offset + length
    return tuple(Part(offset, length) for offset, length in result)


def compute_byte_ranges(parts, method, url, statistics_gatherer):

    if callable(method):
        blocks = method(parts)
    else:
        blocks = parts_heuristics(method, statistics_gatherer)(parts)

    blocks = compress_parts(blocks)

    assert len(blocks) > 0
    assert len(blocks) <= len(parts)

    statistics_gatherer(
        "byte-ranges",
        method=str(method),
        url=url,
        parts=parts,
        blocks=blocks,
    )

    i = 0
    positions = []
    block_offset, block_length = blocks[i]
    for offset, length in parts:
        while offset > block_offset + block_length:
            i += 1
            block_offset, block_length = blocks[i]
        start = i
        while offset + length > block_offset + block_length:
            i += 1
            block_offset, block_length = blocks[i]
        end = i
        # Sanity check: assert that each parts is contain in a rounded part
        assert start == end
        positions.append(
            offset - blocks[i].offset + sum(blocks[j].length for j in range(i))
        )

    return blocks, positions



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import argparse
import cmd
import logging
import os
import sys
import traceback

from termcolor import colored

from .benchmark import BenchmarkCmd
from .cache import CacheCmd
from .check import CheckCmd
from .grib import GribCmd
from .settings import SettingsCmd

try:
    import readline
except ImportError:  # Not availabe on win32

    class readline:
        def set_history_length(*args, **kwargs):
            pass

        def write_history_file(*args, **kwargs):
            pass

        def read_history_file(*args, **kwargs):
            pass


class CliMetLabApp(
    cmd.Cmd,
    SettingsCmd,
    CacheCmd,
    CheckCmd,
    GribCmd,
    BenchmarkCmd,
):
    # intro = 'Welcome to climetlab. Type ? to list commands.\n'
    prompt = colored("(climetlab) ", "yellow")

    rc_file = "~/.climetlab-history"

    def postloop(self):
        try:
            readline.set_history_length(1000)
            readline.write_history_file(os.path.expanduser(self.rc_file))
        except Exception:
            pass

    def preloop(self):
        try:
            readline.read_history_file(os.path.expanduser(self.rc_file))
            with open(os.path.expanduser(self.rc_file), "rt") as f:
                for line in f:
                    self.history.append(line[:-1])
        except Exception:
            pass

    def empty_line(self):
        pass

    def do_quit(self, args):
        """Quit climetlab."""
        return True

    def default(self, line):
        if line == "EOF":
            return True

        cmd = colored(line.split()[0], "yellow")
        help = colored("help", "yellow")

        print(
            f"Unknown command {cmd}. Type {help} for the list of known command names."
        )

    def onecmd(self, line):
        try:
            return super().onecmd(line)
        except ValueError as e:
            print(colored(str(e), "red"))
        except Exception:
            traceback.print_exc()
        return False


def main():
    p = argparse.ArgumentParser()
    p.add_argument(
        "--debug",
        action="store_true",
        help="verbose operation (default: quiet)",
    )
    p.add_argument(
        "cmdline",
        type=str,
        metavar="CMD",
        nargs=argparse.REMAINDER,
    )
    args = p.parse_args()
    sys.argv[1:] = cmdline = args.cmdline

    logging.basicConfig(level=args.debug and "DEBUG" or "WARN")

    app = CliMetLabApp()

    if cmdline:
        return app.onecmd(" ".join(cmdline))
    else:
        app.cmdloop()


if __name__ == "__main__":
    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import cgi
import datetime
import json
import logging
import os
import time

import pytz
import requests
from dateutil.parser import parse as parse_date

from .downloader import Downloader
from .multipart import DecodeMultipart, PartFilter, compute_byte_ranges

LOG = logging.getLogger(__name__)


def NoFilter(x):
    return x


class HTTPDownloaderBase(Downloader):
    supports_parts = True

    def __init__(
        self,
        url,
        verify=True,
        http_headers=None,
        fake_headers=None,
        range_method=None,
        retry_max=500,
        sleep_max=120,
        **kwargs,
    ):
        super().__init__(url, **kwargs)
        self._headers = None
        self._url = None
        self.http_headers = http_headers if http_headers else {}
        self.verify = verify
        self.fake_headers = fake_headers
        self.range_method = range_method
        self.sleep_max = sleep_max
        self.retry_max = retry_max

    def headers(self):
        if self._headers is None or self.url != self._url:
            self._url = self.url
            self._headers = {}
            if self.fake_headers is not None:
                self._headers = dict(**self.fake_headers)
            else:
                try:
                    r = self.robust(requests.head)(
                        self.url,
                        headers=self.http_headers,
                        verify=self.verify,
                        timeout=self.timeout,
                        allow_redirects=True,
                    )
                    r.raise_for_status()
                    for k, v in r.headers.items():
                        self._headers[k.lower()] = v
                    LOG.debug(
                        "HTTP headers %s",
                        json.dumps(self._headers, sort_keys=True, indent=4),
                    )
                except Exception:
                    self._url = None
                    self._headers = {}
                    LOG.exception("HEAD %s", self.url)
        return self._headers

    def extension(self):

        ext = super().extension()

        if ext == ".unknown":
            # Only check for "content-disposition" if
            # the URL does not end with an extension
            # so we avoid fetching the headers unesseraly

            headers = self.headers()

            if "content-disposition" in headers:
                value, params = cgi.parse_header(headers["content-disposition"])
                assert value == "attachment", value
                if "filename" in params:
                    ext = super().extension(params["filename"])

        return ext

    def title(self):
        headers = self.headers()
        if "content-disposition" in headers:
            value, params = cgi.parse_header(headers["content-disposition"])
            assert value == "attachment", value
            if "filename" in params:
                return params["filename"]
        return super().title()

    def transfer(self, f, pbar):
        total = 0
        start = time.time()
        for chunk in self.stream(chunk_size=self.chunk_size):
            self.observer()
            if chunk:
                f.write(chunk)
                total += len(chunk)
                pbar.update(len(chunk))

        self.statistics_gatherer(
            "transfer",
            url=self.url,
            total=total,
            elapsed=time.time() - start,
        )
        return total

    def cache_data(self):
        return self.headers()

    def out_of_date(self, path, cache_data):

        if cache_data is not None:

            # TODO: check 'cache-control' to see if we should check the etag
            if "cache-control" in cache_data:
                pass

            if "expires" in cache_data:
                if cache_data["expires"] != "0":  # HTTP1.0 legacy
                    try:
                        expires = parse_date(cache_data["expires"])
                        now = pytz.UTC.localize(datetime.datetime.utcnow())
                        if expires > now:
                            LOG.debug(
                                "URL %s not expired (%s > %s)", self.url, expires, now
                            )
                            return False
                    except Exception:
                        LOG.exception(
                            "Failed to check URL expiry date '%s'",
                            cache_data["expires"],
                        )

            try:
                headers = self.headers()
            except requests.exceptions.ConnectionError:
                return False

            cached_etag = cache_data.get("etag")
            remote_etag = headers.get("etag")

            if cached_etag != remote_etag and remote_etag is not None:
                LOG.warning("Remote content of URL %s has changed", self.url)
                return True
            else:
                LOG.debug("Remote content of URL %s unchanged", self.url)

        return False

    def check_for_restarts(self, target):
        if not os.path.exists(target):
            return 0

        # Check if we can restarts the transfer
        # TODO: check etags... the file may have changed since

        bytes = os.path.getsize(target)

        if bytes > 0:
            headers = self.headers()
            if headers.get("accept-ranges") != "bytes":
                LOG.warning(
                    "%s: %s bytes already download, but server does not support restarts",
                    target,
                    bytes,
                )
                return 0

            LOG.info(
                "%s: resuming download from byte %s",
                target,
                bytes,
            )

        return bytes

    def issue_request(self, bytes_ranges=None):
        headers = {}
        headers.update(self.http_headers)
        if bytes_ranges is not None:
            headers["range"] = bytes_ranges

        r = self.robust(requests.get)(
            self.url,
            stream=True,
            verify=self.verify,
            timeout=self.timeout,
            headers=headers,
        )
        try:
            r.raise_for_status()
        except Exception:
            LOG.error("URL %s: %s", self.url, r.text)
            raise
        return r

    def robust(self, call):
        def retriable(code):

            return code in (
                requests.codes.internal_server_error,
                requests.codes.bad_gateway,
                requests.codes.service_unavailable,
                requests.codes.gateway_timeout,
                requests.codes.too_many_requests,
                requests.codes.request_timeout,
            )

        def wrapped(*args, **kwargs):
            tries = 0
            while tries < self.retry_max:
                try:
                    r = call(*args, **kwargs)
                except (
                    requests.exceptions.ConnectionError,
                    requests.exceptions.ReadTimeout,
                ) as e:
                    r = None
                    LOG.warning(
                        "Recovering from connection error [%s], attemps %s of %s",
                        e,
                        tries,
                        self.retry_max,
                    )

                if r is not None:
                    if not retriable(r.status_code):
                        return r
                    LOG.warning(
                        "Recovering from HTTP error [%s %s], attemps %s of %s",
                        r.status_code,
                        r.reason,
                        tries,
                        self.retry_max,
                    )

                tries += 1

                LOG.warning("Retrying in %s seconds", self.sleep_max)
                time.sleep(self.sleep_max)
                LOG.info("Retrying now...")

        return wrapped


class FullHTTPDownloader(HTTPDownloaderBase):
    def prepare(self, target):
        assert self.parts is None

        size = None
        mode = "wb"
        skip = 0

        headers = self.headers()
        if "content-length" in headers:
            try:
                size = int(headers["content-length"])
            except Exception:
                LOG.exception("content-length %s", self.url)

        # content-length is the size of the encoded body
        # so we cannot rely on it to check the file size
        trust_size = size is not None and headers.get("content-encoding") is None

        # Check if we can restarts the transfer

        range = None
        bytes = self.check_for_restarts(target)
        if bytes > 0:
            assert size is None or bytes < size, (bytes, size, self.url, target)
            skip = bytes
            mode = "ab"
            range = f"bytes={bytes}-"

        r = self.issue_request(range)

        self.stream = r.iter_content

        LOG.debug(
            "url prepare size=%s mode=%s skip=%s trust_size=%s",
            size,
            mode,
            skip,
            trust_size,
        )
        return (size, mode, skip, trust_size)


class PartHTTPDownloader(HTTPDownloaderBase):
    def prepare(self, target):
        assert self.parts is not None

        headers = self.headers()
        if headers.get("accept-ranges") != "bytes":
            return self.bytes_range_not_supported(target)

        if len(self.parts) == 1:
            return self.one_part_only(target)

        return self.multi_parts(target)

    def bytes_range_not_supported(self, target):
        LOG.warning(
            "Server for %s does not support byte ranges, downloading whole file",
            self.url,
        )

        request = self.issue_request()
        self.stream = PartFilter(self.parts)(request.iter_content)

        size = sum(p.length for p in self.parts)
        return (size, "wb", 0, True)

    def one_part_only(self, target):
        # Special case, we let HTTP to its job, so we can resume transfers if needed
        assert len(self.parts) == 1

        offset, length = self.parts[0]
        start = offset
        end = offset + length - 1
        bytes = self.check_for_restarts(target)
        if bytes > 0:
            start += bytes
            skip = bytes
            mode = "ab"
        else:
            skip = 0
            mode = "wb"

        bytes_range = f"bytes={start}-{end}"
        request = self.issue_request(bytes_range)
        self.stream = request.iter_content

        size = sum(p.length for p in self.parts)
        return (size, mode, skip, True)

    def split_large_requests(self, parts):
        ranges = []
        for offset, length in parts:
            ranges.append(f"{offset}-{offset+length-1}")

        # Nginx default is 4K
        # https://stackoverflow.com/questions/686217/maximum-on-http-header-values
        bytes_range = f"bytes={','.join(ranges)}"

        if len(bytes_range) <= 4000:
            return [(bytes_range, parts)]

        middle = len(parts) // 2
        return self.split_large_requests(parts[:middle]) + self.split_large_requests(
            parts[middle:]
        )

    def multi_parts(self, target):

        # TODO: implement transfer restarts by trimming the list of parts

        filter = NoFilter
        parts = self.parts

        if self.range_method:

            rounded, positions = compute_byte_ranges(
                self.parts,
                self.range_method,
                self.url,
                self.statistics_gatherer,
            )
            filter = PartFilter(self.parts, positions)
            parts = rounded

        splits = self.split_large_requests(parts)

        def iterate_requests(chunk_size):

            for bytes_ranges, parts in splits:

                request = self.issue_request(bytes_ranges)

                stream = DecodeMultipart(
                    self.url,
                    request,
                    parts,
                    verify=self.verify,
                    timeout=self.timeout,
                    headers=self.http_headers,
                )

                yield from stream(chunk_size)

        self.stream = filter(iterate_requests)

        size = sum(p.length for p in self.parts)
        return (size, "wb", 0, True)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
import os
from ftplib import FTP
from urllib.parse import urlparse

from .downloader import Downloader

LOG = logging.getLogger(__name__)


class FTPDownloaderBase(Downloader):

    supports_parts = False

    def __init__(self, url, **kwargs):
        super().__init__(url, **kwargs)

    def prepare(self, target):

        o = urlparse(self.url)
        assert o.scheme == "ftp"

        if "@" in o.netloc:
            auth, server = o.netloc.split("@")
            user, password = auth.split(":")
        else:
            auth, server = None, o.netloc
            user, password = "anonymous", "anonymous"

        ftp = FTP(
            server,
            timeout=self.timeout,
        )

        if auth:
            ftp.login(user, password)
        else:
            ftp.login()

        ftp.cwd(os.path.dirname(o.path))
        ftp.set_pasv(True)
        self.filename = os.path.basename(o.path)
        self.ftp = ftp

        return (ftp.size(self.filename), "wb", 0, True)

    def transfer(self, f, pbar):
        total = 0

        def callback(chunk):
            nonlocal total
            self.observer()
            f.write(chunk)
            total += len(chunk)
            pbar.update(len(chunk))

        self.ftp.retrbinary(f"RETR {self.filename}", callback)

    def finalise(self):
        self.ftp.close()


class FullFTPDownloader(FTPDownloaderBase):
    pass


class PartFTPDownloader(FTPDownloaderBase):
    def __init__(self, url, **kwargs):
        # If needed, that can be implemented with the PartFilter
        raise NotImplementedError("Part FTPDownloader is not yet implemented")



#!/usr/bin/env python3
# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import argparse

from climetlab.scripts.main import CliMetLabApp


def execute(*args):

    for cmd in sorted([c for c in dir(CliMetLabApp) if c.startswith("do_")]):
        func = getattr(CliMetLabApp, cmd)
        if not hasattr(func, "_argparser"):
            continue

        cmd = cmd[3:]  # remove "do_"

        print()
        print(f".. _{cmd}-command:")
        print()
        print(cmd)
        print("^" * len(cmd))
        print()

        # Not garanteed to work with future versions of Python
        func._argparser.formatter_class = (
            lambda prog: argparse.RawDescriptionHelpFormatter(
                prog,
                width=90,
                max_help_position=10000,
            )
        )

        for n in func._argparser.format_help().split("\n"):

            if n.startswith("usage: "):
                print()
                print(".. code-block:: text")
                print()
                print(" ", n)
                continue

            if len(n.strip()) == 0:
                print()
                print()
                continue

            if n.endswith(" arguments:"):
                print(n[0].upper() + n[1:])
                print('"' * len(n))
                print()
                print(".. code-block:: text")
                print()
                continue

            if n.strip().startswith("-"):
                print(" ", n)
                continue

            print(n)


if __name__ == "__main__":
    execute()



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import json
import os

import climetlab as cml


class IndexBackend:
    pass


class JsonIndexBackend(IndexBackend):
    def __init__(self, filename):
        self._entries = None
        self.filename = filename

    @property
    def entries(self):
        if self._entries is not None:
            return self._entries

        self._entries = []

        if os.path.exists(self.filename):
            self.read_file_index(self.filename)
            return self._entries

        if self.filename.startswith("http://") or self.filename.startswith("https://"):
            path = cml.load_source("url", self.filename).path
            self.read_file_index(path)
            return self._entries

        raise Exception(f"Cannot load index data from {self.filename}")

    def read_file_index(self, filename):
        with open(filename, "r") as f:
            lines = f.readlines()
        for line in lines:
            entry = json.loads(line)
            self._entries.append(entry)

    def match(self, entry, request):
        for k, v in request.items():
            if isinstance(v, (tuple, list)):
                if not entry[k] in v:
                    return False
            else:
                if entry[k] != v:
                    return False
        return True

    def lookup(self, request):
        parts = []
        for e in self.entries:
            if self.match(e, request):
                path = e.get("_path", None)
                offset = int(e["_offset"])
                length = int(e["_length"])
                parts.append((path, [offset, length]))
        return parts



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import sys
from urllib.parse import urlparse

from .downloader import Downloader

LOG = logging.getLogger(__name__)


class FileDownloaderBase(Downloader):

    supports_parts = True

    def __init__(self, url, **kwargs):
        super().__init__(url, **kwargs)
        o = urlparse(self.url)
        path = o.path

        if sys.platform == "win32" and self.url.startswith("file://"):
            # this is because urllib does not decode
            # 'file://C:\Users\name\climetlab\docs\examples\test.nc'
            # as expected.
            path = self.url[len("file://") :]

        if sys.platform == "win32" and path[0] == "/" and path[2] == ":":
            path = path[1:]

        self.path = path


class FullFileDownloader(FileDownloaderBase):
    def local_path(self):
        return self.path


class PartFileDownloader(FileDownloaderBase):
    def prepare(self, target):
        parts = self.parts
        size = sum(p.length for p in parts)
        return (size, "wb", 0, True)

    def transfer(self, f, pbar):
        with open(self.path, "rb") as g:
            total = 0
            for offset, length in self.parts:
                g.seek(offset)
                self.observer()
                while length > 0:
                    chunk = g.read(min(length, self.chunk_size))
                    assert chunk
                    f.write(chunk)
                    length -= len(chunk)
                    total += len(chunk)
                    pbar.update(len(chunk))
        return total



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab import load_source
from climetlab.indexing import PerUrlIndex
from climetlab.sources.multi import MultiSource

from .prompt import APIKeyPrompt


class EODKeyPrompt(APIKeyPrompt):
    register_or_sign_in_url = "https://www.ecmwf.int/"
    retrieve_api_key_url = "https://www.ecmwf.int"

    prompts = [
        dict(
            name="url",
            default="c",
            title="API url",
            validate=r"http.?://.*",
        ),
    ]

    rcfile = "~/.ecmwf-open-data"


class EODRetriever(MultiSource):

    sphinxdoc = """
    EODRetriever
    """

    def __init__(self, *args, **kwargs):
        if len(args):
            assert len(args) == 1
            assert isinstance(args[0], dict)
            assert not kwargs
            kwargs = args[0]

        prompt = EODKeyPrompt()
        self.config = prompt.check(load=True)

        options = dict(
            url=self.config["url"],
            date=-1,
            step=0,
            resol="0p4",
            stream="oper",
            type="fc",
            extension=".grib2",
        )
        options.update(kwargs)
        # assert False, (kwargs, options)

        sources = self.requests(**options)

        super().__init__(sources)

    # @normalize("date", "date-list(%Y-%m-%d)")
    # @normalize("area", "bounding-box(list)")
    def requests(self, **request):

        for k, v in request.items():
            if not isinstance(v, (list, tuple)):
                request[k] = [v]

        pattern = (
            "{url}/{date:date(%Y%m%d)}/{date:date(%H)}z/{resol}/{stream}/"
            "{date:date(%Y%m%d%H%M%S)}-{step}h-{stream}-{type}{extension}"
        )

        index = PerUrlIndex(pattern, substitute_extension=True)

        # result = []
        # for p in Pattern(pattern).substitute([], **kwargs):
        #    result.append(load_source("url", p, lazily=True))# parts=[(0, 4)]))

        sources = load_source("indexed-urls", index, request)

        return sources


source = EODRetriever



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import re
from collections import namedtuple

LOG = logging.getLogger(__name__)


Part = namedtuple("Part", ["offset", "length"])


def round_down(a, b):
    return (a // b) * b


def round_up(a, b):
    return ((a + b - 1) // b) * b


class HierarchicalClustering:
    def __init__(self, min_clusters=5):
        self.min_clusters = min_clusters

    def __call__(self, parts):
        clusters = [p for p in parts]

        while len(clusters) > self.min_clusters:
            min_dist = min(
                clusters[i].offset - clusters[i - 1].offset + clusters[i - 1].length
                for i in range(1, len(clusters))
            )
            i = 1
            while i < len(clusters):
                d = clusters[i].offset - clusters[i - 1].offset + clusters[i - 1].length
                if d <= min_dist:
                    clusters[i - 1] = Part(
                        clusters[i - 1].offset,
                        clusters[i].offset
                        + clusters[i].length
                        - clusters[i - 1].offset,
                    )
                    clusters.pop(i)
                else:
                    i += 1

        return clusters

    def __repr__(self):
        return f"cluster({self.min_clusters})"


class BlockGrouping:
    def __init__(self, block_size):
        self.block_size = block_size

    def __call__(self, parts):
        blocks = []
        last_block_offset = -1
        last_offset = 0

        for offset, length in parts:

            assert offset >= last_offset

            block_offset = round_down(offset, self.block_size)
            block_length = round_up(offset + length, self.block_size) - block_offset

            if block_offset <= last_block_offset:
                prev_offset, prev_length = blocks.pop()
                end_offset = block_offset + block_length
                prev_end_offset = prev_offset + prev_length
                block_offset = min(block_offset, prev_offset)
                assert block_offset == prev_offset
                block_length = max(end_offset, prev_end_offset) - block_offset

            blocks.append(Part(block_offset, block_length))

            last_block_offset = block_offset + block_length
            last_offset = offset + length

        return blocks

    def __repr__(self):
        return f"blocked({self.block_size})"


class Automatic:
    def __call__(self, parts):
        smallest = min(x.length for x in parts)
        range_method = round_up(max(x.length for x in parts), 1024)

        while range_method >= smallest:
            blocks = BlockGrouping(range_method)(parts)
            range_method //= 2

        # Max number of parts
        return blocks

    def __repr__(self):
        return "auto"


class Sharp:
    def __init__(self, transfer_rate=1024 * 1024, request_latency_overhead=0.1):
        """
        transfer_rate:
            unit is byte/seconds.
            additional cost to download one useles byte.
         request_latency_overhead:
            unit is seconds.
            additional cost to do one additional HTTP request.
        """
        self.transfer_rate = transfer_rate
        self.request_latency_overhead = request_latency_overhead

    def __call__(self, parts):

        blocks = [parts[0]]

        for offset, length in parts[1:]:
            latest = blocks[-1]
            latest_end = latest.offset + latest.length

            distance = offset - latest_end
            assert distance >= 0, (distance, latest, (offset, length), parts)

            cost_of_merging = (
                distance / self.transfer_rate - self.request_latency_overhead
            )

            if cost_of_merging <= 0:
                blocks[-1] = Part(latest.offset, latest.length + length)
            else:
                blocks.append(Part(offset, length))

        return blocks

    def __repr__(self):
        return f"sharp({self.transfer_rate},{self.request_latency_overhead})"


class Pipe:
    def __init__(self, first, second):
        self.first = first
        self.second = second

    def __call__(self, parts):
        return self.first(self.second(parts))

    def __repr__(self):
        return f"{self.second}|{self.first}"


class Debug:
    def __call__(self, parts):
        print("DEBUG", parts)
        return parts

    def __repr__(self):
        return "debug"


HEURISTICS = {
    "auto": Automatic,
    "sharp": Sharp,
    "cluster": HierarchicalClustering,
    "blocked": BlockGrouping,
    "debug": Debug,
}


def parts_heuristics(method, statistics_gatherer):
    if isinstance(method, int):
        return BlockGrouping(method)

    result = None
    for name in method.split("|"):
        if "(" in name:
            m = re.match(r"(.+)\((.+)\)", name)
            name = m.group(1)
            args = []
            for a in m.group(2).split(","):
                try:
                    args.append(int(a))
                except ValueError:
                    args.append(float(a))
        else:
            args = []

        obj = HEURISTICS[name](*args)
        if result is None:
            result = obj
        else:
            result = Pipe(obj, result)

    statistics_gatherer(
        "parts-heuristics",
        full_method=str(method),
        method_args=args,
    )
    return result



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import mimetypes
import os
from urllib.parse import urlparse

from .file import FullFileDownloader, PartFileDownloader
from .ftp import FullFTPDownloader, PartFTPDownloader
from .heuristics import Part
from .http import FullHTTPDownloader, PartHTTPDownloader
from .multipart import compress_parts

LOG = logging.getLogger(__name__)


def mimetype_to_extension(mimetype, compression, default=".unknown"):
    EXTENSIONS = {
        None: "",
        "gzip": ".gz",
        "xz": ".xz",
        "bzip2": ".bz2",
    }
    if mimetype is None:
        extension = default
    else:
        extension = mimetypes.guess_extension(mimetype)

    return extension + EXTENSIONS[compression]


def canonical_extension(path):
    _, ext = os.path.splitext(path)
    ext = mimetype_to_extension(*mimetypes.guess_type(path), default=ext)
    # Looks like mimetypes as .cdf before .nc
    # TODO: report it to Python's bug tracker
    EXTENSIONS = {".cdf": ".nc"}

    return EXTENSIONS.get(ext, ext)


DOWNLOADERS = {
    ("ftp", False): FullFTPDownloader,
    ("ftp", True): PartFTPDownloader,
    ("http", False): FullHTTPDownloader,
    ("http", True): PartHTTPDownloader,
    ("https", False): FullHTTPDownloader,
    ("https", True): PartHTTPDownloader,
    ("file", False): FullFileDownloader,
    ("file", True): PartFileDownloader,
}


def get_downloader(url, **kwargs):

    parts = kwargs.get("parts")
    if parts is not None:
        parts = [Part(offset, length) for offset, length in parts]
        parts = compress_parts(parts)
        if len(parts) == 0:
            parts = None
        kwargs["parts"] = parts

    o = urlparse(url)
    has_parts = parts is not None and len(parts) > 0

    downloader = DOWNLOADERS[(o.scheme, has_parts)](url, **kwargs)

    return downloader



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os

LOG = logging.getLogger(__name__)


def _ignore(*args, **kwargs):
    pass


class NoBar:
    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        pass

    def close(self, *args, **kwargs):
        pass


class Downloader:
    def __init__(
        self,
        url,
        chunk_size=1024 * 1024,
        timeout=None,
        parts=None,
        observer=_ignore,
        statistics_gatherer=_ignore,
        progress_bar=NoBar,
        **kwargs,
    ):
        self.url = url
        self.chunk_size = chunk_size
        self.timeout = timeout
        self.parts = parts
        self.observer = observer
        self.statistics_gatherer = statistics_gatherer
        self.progress_bar = progress_bar

    def local_path(self):
        return None

    def extension(self, filename=None):
        """Return extensions:
        - from filename if not None.
        - else from self.url
        - else return ".unknonwn"
        """
        extensions = []
        if filename:
            extensions = self._extensions(filename)
        if not extensions:
            url_no_args = self.url.split("?")[0]
            extensions = self._extensions(url_no_args)
        if not extensions:
            extensions.append(".unknown")
        return "".join(reversed(extensions))

    def _extensions(self, url_or_filename):
        base = os.path.basename(url_or_filename)
        extensions = []
        while True:
            base, ext = os.path.splitext(base)
            if not ext:
                break
            extensions.append(ext)
        return extensions

    def download(self, target):
        if os.path.exists(target):
            return

        download = target + ".download"
        LOG.info("Downloading %s", self.url)

        size, mode, skip, trust_size = self.prepare(download)

        with self.progress_bar(
            total=size,
            initial=skip,
            desc=self.title(),
        ) as pbar:

            with open(download, mode) as f:
                total = self.transfer(f, pbar)

            pbar.close()

        if trust_size and size is not None:
            assert (
                os.path.getsize(download) == size
            ), f"File size mismatch {os.path.getsize(download)} bytes instead of {size}"

        os.rename(download, target)

        self.finalise()
        return total

    def finalise(self):
        pass

    def title(self):
        return os.path.basename(self.url)

    def cache_data(self):
        return None

    def out_of_date(self, path, cache_data):
        return False
